\ Clean Songlist - for the Digital Piano -- include 3iPiano.fs  

\ TODO: ADD REST NOTATIONS
\ TODO: ADD variable tones (half and sixteenth notes)
\ TODO: FIX - leave doesn't work
\ Design: this plays by reading notes or directly e.g. 1 .. 2 .. 3 ..

: include-song-files
  s" 3iForth.fs"        included  500 ms
  s" 3iSound-4x4.fs"    included  500 ms  \ includes 3iMusic-scales
  s" 3iShift.fs"        included  500 ms
  s" 3iPianoLights.fs"  included  500 ms
\  s" 3iArpeggiator.fs"  included  500 ms
;

include-song-files  \ includes 3iSongs, 3iMusic-menu

\ MUSICAL NOTATION FOR SONGLIST.fs
\ Design: SONG GRAMMAR consists of (1) 8 diatonic notes 1..8,  (2) 6 chromatic notes 9-14, (3) octave down/up, (4) musical rest short/long
\ Design: the grammar allows writing music in easy-to-read 1-8 notation, #/% for sharp/flat, <</>> for octave down/up, and //,///,//3,//// for short/long musical rest
\ Design: scaleroot sets default index to middle c  (c_ / 36 note or 262 Hz tone), span c_ to c'
\ Design: scaleroot gets adjusted to move virtually between octaves, avoiding physically moving these
variable scaleroot 
: scaleroot!      ( note-name -- ) scaleroot !    ;
: scaleroot@      ( -- note-name ) scaleroot @    ;
: reset-scaleroot ( -- )           c_  scaleroot! ;
reset-scaleroot

\ Design: 1 2 ... 8 notation for white keys indexed to middle c_
\ Design: Sharps and Flats added through decorating 1..8
\ Design: cmaj-map maps 1...8 (white keys) and 9..14 (black keys) to the sequential piano note index
9  constant 1#
9  constant 2%
10 constant 2#
10 constant 3%
11 constant 4#
11 constant 5%
12 constant 5#
12 constant 6%
13 constant 6#
13 constant 7%
14 constant 8#

\ CODE REVIEW: use >mem
\ map: diatonic to chromatic, i.e. allows use of 1..8
create cmaj-map 14 bytes allot
white-registers black-registers 13 14 cmaj-map >mem drop
: [...]  ( playlist# -- 0index# )     1 - cmaj-map + c@ ;  \ transforms playlist notation to sequential notes 0-indexed

\ CODE REVIEW: what is this for?  reading notes from device?  but then change so not have first black key on 10, but rather 9, and do away with this...
create note-to-0index 14 bytes allot
white-registers -1 black-registers 14 note-to-0index >mem drop 
: translatenote ( recordnote -- 0indexnote )  \ takes songlist grammar note [diatonic] (1..14) and converts to piano sequence (0..13) [chromatic]
    1 - bytes note-to-0index + c@ ;
\ : test 15 1 do i dup . translatenote dup . _soundlight 200 ms cr loop ;

\ Design: this is the inverse map: chromatic to diatonic.  Used for?
create chr2cmaj-map  14 bytes allot
1 9 2 10 3 4 11 5 12 6 13 7 8 14  14 chr2cmaj-map >mem drop
: chr2dia ( 0indexnote -- grammarnote )  \ takes 0index chromatic note and converts to 1-index cmaj note
    bytes chr2cmaj-map + c@ ;
\ : test 13 0 do i dup . chr2dia . cr loop ;

\  Design: on 32-key F__ to C'' keyboard seq# is -7 to 24  (piano-led board)
: [,,,]  ( 1-8# or 9-14# -- 0index# ) 1 - scale + c@ c_ - ;   \ provides relative index +/- of note referenced to c_ middle c for transposition

\ Design: non-note grammar are above the c.96 notes in the piano scale (extended manually to custom frequencies)
\ Design: virtual (software only) octave adjustments.  See ><< and >>> in 3iSound-4x4.fs for physical/hardware octave adjustments in buttons
$FA constant <<  \ octave down
$FB constant >>  \ octave up
: is-octave? ( g -- g 0/1 ) dup << >> between? ;
: [<<] ( -- ) scaleroot@ 12 - scaleroot! ;   \ lower octave
: [>>] ( -- ) scaleroot@ 12 + scaleroot! ;   \ raise octave
: adjust-octave ( g:<<,>> --  ) >> - if [<<] else [>>] then ;

\ Design: musical rests
$FC constant //    \ short rest R1
$FD constant ///   \       rest R2
$FE constant 3///  \       rest R3
$FF constant 4///  \ long  rest R4
create rests ' R4 , ' R3 , ' R2 , ' R1 ,
: is-rest? ( g -- g 0/1 ) dup // 4/// between? ;
: musical-rest ( g --  ) $FF swap - cells rests + @ execute ;

\ Design: storage for SONG GRAMMAR g for playback, transpositions etc.
create songmem 512 bytes allot
variable songptr
variable songlen
: songlen@ songlen @ ;
: reset-songlen   ( -- )   0 songlen ! ;
: fresh-song      ( -- )   songmem songptr !  reset-songlen ;
: songlen++       ( -- )   songlen@ 1 + songlen ! ;
: songc!+         ( g -- ) songptr @   c!+  songptr ! songlen++ ;
: song-mem2list   ( n -- list n )  \ memory-array to list-stack
    >r songmem r@ 0 do c@+ swap loop drop r> ;
: showsongmem     ( -- ) songlen@ songmem over 0 do c@+ . loop drop ;  \ dump-memory
: showsonglist    ( n -- ) dup 0 do dup i - roll . loop drop ;  \ dump-list
fresh-song

\ OBSERVATION:  5flats  lowered 1 semitone gives a nice cmaj-white keys scale!

: pausedur 500 ms ;

: _key              ( -- 0/1 )    key esc-key = if leave then ;  \ BUG!!! esc-key needs to trigger leaving outerloop
: _sound            ( note# -- )  note ;            \ transform to note# (3iSound-4x4.fs) and plays note
: [light]           c_ - 8 + ;    \ calibrated to middle C
: _light            ( note# -- )  [light] kk ;     \ transform to light# (3iPianoLights.fs) and turns on light
: _soundlight       ( note# -- )  dup _light _sound ;
: _soundlightkey    ( note# -- )  dup . _soundlight _key ; \ advance on keypress, ESC to stop
: _soundlightbutton ( note# -- )  
    begin dup _soundlight poll-note-data translatenote scaleroot@ + 2dup = 
        if drop pausedur exitloop 
        else drop 32-clear pausedur pausedur continueloop 
        then 
    until drop ;
: _lightkey         ( 0index# -- )  _light _key ;
: _soundkey         ( 0index# -- )  _sound _key ;

variable playmode
0 constant  soundonly
1 constant  lightonly
2 constant  soundlight
3 constant  soundlightkey
4 constant  lightkey
5 constant  soundkey
6 constant  soundlightbutton
: playmode!  ( choice -- ) playmode ! ;
: soundonly-mode  soundonly playmode! ;
: soundlight-mode soundlight playmode! ;
: soundlightkey-mode soundlightkey playmode! ;
: soundlightbutton-mode soundlightbutton playmode! ;
: soundkey-mode soundkey playmode! ;
: lightkey-mode lightkey playmode! ;
soundonly-mode  \ default at startup

: playmode@  ( -- choice ) playmode @ ;
create playmode-map   ' _sound , ' _light , ' _soundlight , ' _soundlightkey , ' _lightkey , ' _soundkey , ' _soundlightbutton ,
  \ Design: .. doesn't always sound a note, if << or >> then it adjusts the octave and stops as the next note needs to be rolled in from the stack
: do-playmode ( note# -- ) playmode-map playmode@ cells + @ execute ;

\ CODE REVIEW: Get to a musical data structure -- 3 arrays
\ play directly from stack, left to right


\ ================= ALGORITHM -- EXTRACT / DISPLAY KEY SIGNATURE FROM SONG
\ \ Design: extract uniques
create songuniq 45 bytes allot \ enough to surround the 32-key pianolights board
variable uniqptr
variable uniqlen
: reset-uniqptr  songuniq uniqptr !   0 uniqlen ! ;
: uniqptr@   uniqptr @ ; 
: uniqptr++  uniqptr@ 1 + uniqptr ! ;
: uniqlen@  uniqlen @ ;
: uniqlen++  uniqlen @ 1 + uniqlen ! ;
: add-uniq ( x -- ) uniqptr@ ! uniqptr++ uniqlen++  ;  \ ignore non-note grammars
: [show-uniq] ( -- ) uniqlen@ 0 > if songuniq uniqlen@ 0 do c@+ . loop drop else ." empty " then ;

: end?  ( uptr -- 1/0 ) dup uniqptr@ = ;
: uniq? ( x uptr -- x uptr' 1/0 ) over swap c@+ rot = ;  \ CODE REVIEW: this is actually 0 for uniq, 1 for not uniq --- misnamed
: search-uniq ( x uptr -- 0/1 )     
    begin uniq?
        if  \ not-unique, i.e. matching entry found
            drop exitloop
        else \ uniq, i.e. entry not found, continue search unless at end of list
            end? 
                if \ at end of list
                    swap add-uniq exitloop 
                else continueloop 
                then 
        then 
    until
    drop
;
: extract-uniqs ( songlen -- uniqlen ) \ 
     reset-uniqptr songmem swap 0 
     do \ loop over song grammar elements
        c@+ dup << < 
            if songuniq end?   \ found note
                if drop add-uniq \ reached end of list, so add
                else search-uniq
                then
            else drop \ ignore non-notes
            then 
    loop
    drop 
    uniqlen@
;


: play-uniqs ( uniqlen -- uniqlen )
  songuniq over 0 do c@+ note R2 loop drop ;
\ CODE REVIEW: need ORDER algorithm
: [uniqlights] ( uniqlen -- list uniqlen ) 
    >r songuniq r@ 0 do c@+ [light] swap loop drop r> ;
: uniqlights ( uniqlen -- ) [uniqlights] lightn ;
: songlights ( songlen -- songlen ) dup extract-uniqs uniqlights ;

\ TODO -- translate to light IDs and then light them!
\ ================= END UNDER CONSTRUCTION


\ Design: Takes song grammar from stack and writes notes to memory FIXED WRT TO A SCALE and INTERPRETING virtual octaves, adding rests grammar.  
\             Example: dream [writesong] play
: [writenote]     ( g -- )
        is-octave? 
        if   adjust-octave     \ CODE REVIEW: add light for adjusting octave
        else is-rest?
              if  ( pass through )
              else [,,,] scaleroot@ + then
              songc!+  \ CODE REVIEW: count use of the array so know the length
        then 
;
: [writesong]     ( <1..8# notes> n -- ) 
    reset-scaleroot fresh-song  dup 0 
    do 
        dup i - roll 
        [writenote]
    loop drop ;

: list-uniqs ( <list> songlen -- uniqlen ) [writesong] extract-uniqs [show-uniq] uniqlen@ dup . ;

\   Design: songlist puts the entire song notation onto the stack and then reads from stack left to right  CODE REVIEW: store into memory, read from memory
\   Design: core musical dependency is 3iSound-4x4.fs which has the notational transformation required
\   Design: audio only
: .. ( grammar -- )  \ notation is 3iSonglist grammar: 1..8 2% 2# >> << // /// //3 ////  CX!!! in pace mode, extra delay because // pauses are read!
    is-rest?  if musical-rest
             else do-playmode
             then ;
: flashlights  32-on 300 ms 32-clear 200 ms ;

\ Design: transposition is done from note index space to avoid having to translate back and forth to songlist grammer (user notation space) which would change length of songmem as octave movements would be different
\ Design: only transpose notes (not rests).  Transposition is 1-1 and cells are overwritten with results, songlen unchanged
: c<!+ ( x ptr -- ptr ) 1 - c!+ ;  \ overwrite previous character with x
: [ntranspose] ( semi# -- ) set-semi# songlen@ songmem over 0 
    do 
        c@+   dup is-rest? if ( ignore ) else get-semi# + then swap c<!+ 
    loop 2drop ;

\ Design: these actions work on song notes and rests stored in memory
: replay      ( -- ) songlen@ songmem over 0 do c@+ .. loop 2drop ;  \ repeats song in memory
: [showuniqs] ( -- ) songlen@ dup extract-uniqs uniqlights drop ;
: ntranspose  ( #steps -- ) [ntranspose] [showuniqs] ;
: transpose   ( -- ) 1 ntranspose ; \ single-step transpose

\ TODO: to get back to user space, rebaseline (song-delta) against c_ (middle c) and use the chr2dia inverse map !!


\ TODO: calculate new 1..8 notation with the revised scale...
\ TODO: algorithm to choose transposition that gives optimal fingering, e.g. c major 
\ TODO: illuminate to show all the (uniq) notes of a scale... to see if one likes the fingering -- DONE
\ TODO: store songs in EEPROM (non-volatile) for playing/replaying rapidly - may lead to timing issues
\ TODO: scan program and highlight all keys to expose the underlying scale based on keys requested
\ : transpose-test ( g -- )  \ shows 7 steps of transformation between notations.   5flats  1 6 5 7 -->  6 11 3 5
\    cr dup .  [,,,] dup . scaleroot@ + dup . 1 + dup . scaleroot@ - dup . 12 mod dup . chr2dia dup . drop ;

\ Design: Song user interface to entry is 1..8,//,///,//3,////,>>,<< grammer loaded onto stack, written to memory, play, transpose, analysis is from memory
\ Design: songlen@ holds length of song (forming object with songmem)
\ Design: these actions work on song in list
: play        ( <songlist> -- )    soundonly-mode        [writesong] replay ;
: watch       ( <songlist> -- )    soundlight-mode       [writesong] replay ;
\ TODO - 2021-09-04---Sat---10:41 when pacing, show the key 1..8 (user space), not the note
: pace        ( <songlist> -- )    soundlightkey-mode    [writesong] replay ; \ learns the pace to play it at, keyboard control

: victory-jingle  80 setbeat  soundonly-mode  1 3 5 8 // 2 4 6 >> 2 <<   11 play 120 setbeat ;
: well-done!   flashlights flashlights  victory-jingle ;

: learn       ( <songlist> -- )    soundlightbutton-mode [writesong] replay  well-done! ; 
: listenstep  ( <songlist> -- )    soundkey-mode         [writesong] replay ;
: lightstep   ( <songlist> -- )    lightkey-mode         [writesong] replay ;
: showuniqs   ( list songlen -- )                        [writesong] [showuniqs] ; 


\ interesting: starting on E makes a lot of complicated songs easy!  dream -1   stairway 4
\ TODO: Add a "next note" button in case the learn mode gets stuck...

\   Design: audio + lights
\  : watch ( n -- ) dup 0 do dup i - roll dup #_ .. loop drop 32-clear ;   \ see 3iPianoLights.fs for kk
\  : watch ( n -- ) dup 0 do dup i - roll #__ 500 ms light| 100 ms loop drop light| ;   \ see 3iPort-4x4.fs for light_

\   Design: learn each note and use PC to prompt next note --- LATER: read key to check it's the right one being pressed
\ : learn ( n -- ) dup 0 do dup i - roll dup #_ .. key drop loop drop 32-clear ;
\ : learn ( n -- ) reset-scaleroot dup 0 do dup i - roll .. key drop loop drop 32-clear ;

\ TO DO: add ESC to play rest.  Add back key to go back.  Add play a phrase key.
\ TO DO: how to flash?


\ see SongList for map to musical notes and multiple keys

\ chords
\ Main chord types (maj, min, dim, aug): https://www.schoolofrock.com/resources/keyboard/piano-chords-for-beginners
\ Triads: https://en.wikipedia.org/wiki/Triad_(music)
\ Sevenths: https://www.musictheory.net/lessons/45
\ Ninths: https://blog.discmakers.com/2020/12/ninth-chords-add-ambiance/
\ Elevenths: https://blog.discmakers.com/2021/03/eleventh-chords/
\ Arpeggio: https://en.wikipedia.org/wiki/Arpeggio

variable rt  \ root note
: rt! ( key-id -- key-id ) dup rt c! ;
: rt@ ( -- key-id ) rt c@ ;
variable shp \ chord shape
: shp! ( shp-id -- ) shp c! ;
: shp@ ( -- shp-id ) shp c@ ;
variable chrdlen
: chrdlen! ( x -- ) chrdlen ! ;
: chrdlen@ ( -- x ) chrdlen @ ;

0  constant majshp
1  constant minshp
2  constant dimshp
3  constant augshp
4  constant sus4shp
5  constant sus2shp
6  constant dom7shp
7  constant maj7shp
8  constant min7shp
9  constant hdim7shp \ half-diminished 7th
10 constant dim7shp  \ fully-diminished 7th

\ major chord shape: root + 4 + 3
\ Design: this works for notes and lights.  Need to bypass the play, and go straight to reading from songmem

: major  ( root -- a b c )  dup 4 + dup 3 +   3 chrdlen! ;
: minor  ( root -- a b c )  dup 3 + dup 4 +   3 chrdlen! ;
: dimin  ( root -- a b c )  dup 3 + dup 3 +   3 chrdlen! ;
: augmen ( root -- a b c ) dup 4 + dup 4 +    3 chrdlen! ;
: sus4_  ( root -- a b c )  dup 5 + dup 2 +   3 chrdlen! ;
: sus2_  ( root -- a b c )  dup 2 + dup 5 +   3 chrdlen! ;
: domin7_  ( root -- a b c d ) dup 4 + dup 3 + dup 3 +   4 chrdlen! ; \ dominant 7th
: major7 ( root -- a b c d ) dup 4 + dup 3 + dup 4 +   4 chrdlen! ;
: minor7 ( root -- a b c d ) dup 3 + dup 4 + dup 3 +   4 chrdlen! ;
: halfdim7_  ( root -- a b c d ) dup 3 + dup 3 + dup 4 +   4 chrdlen! ;
: dimin7 ( root -- a b c d ) dup 3 + dup 3 + dup 3 +   4 chrdlen! ;

create chord-exec ' major , ' minor , ' dimin , ' augmen , ' sus4_ , ' sus2_ , ' domin7_ , ' major7 , ' minor7 , ' halfdim7_ , ' dimin7 ,  
: runchord (  -- ) shp@ 4 * chord-exec + @ execute ;
: light-chord ( note -- ) [light] rt! runchord chrdlen@ lightn ; 
\ Design: uses separate memory for chords (chordmem) and songs (songmem)
\ Design: plays root also at end of triad for musicality
create chordmem 9 bytes allot
: chordplay   ( -- ) chordmem chrdlen@ 0 do c@+ .. loop drop ; \ plays chord in memory  CODE REVIEW: factor with replay
\ Design: plays 2nd, 3rd, and 1st, as initial press is for the first   \ TODO: add silent mode to .avr 
: sound-chord ( note -- ) runchord chrdlen@ chordmem >mem drop chordplay ;
: playchord ( note -- note ) dup light-chord dup sound-chord ;
: testchord ( note -- ) playchord . 2 songmem .mem ."  | " shp@ . ;
  
\ TODO: menu to choose chord shape/type  (e.g. x shp!)
\ TODO - design fast sweep of chords, and chord-mode option in Songlist.fs
\ CODE REVIEW: should be [,,] instead of translatenote (fix it in the .avr listing the keyprpesses)
\ extract-note returns 1--8 white keys and 10-14 black keys
\ : fix-defect ( extract-note ) dup 9 > if 1 - then ;  \ needed to fix previous key mapping 1-8, 10-14
: chordator ( -- ) \ NOTE: white keys only!
  1 silent 
  beat >r 70 setbeat  begin   
    fetch-note-data 0 > if  extract-note  [,,] playchord . then key?  \ BUG: why does extract-note give 0 output sometimes
  until 32-clear r> setbeat 0 silent ;
: "" ( rootnote -- ) dup is-rest? if musical-rest else [,,] playchord drop then ;

\ TODO: this provides a nice record notes functionality!  Now add edit notes


\ Design: store arpeggiator to loop over chords, 4 bars, each bar is 16 semiquavers, so 64 slots
create arpeg 64 bytes allot
: >narpeg ( <arpeg> n -- n ) arpeg >mem ;
\ : arpeg-chord 

\ ex. c-maj = C E G ;  \ ex. c-maj7 = C E G B ;

\ TODO - add arpeggio structure eg //1353//
\ TODO - add note names for chords... instead of sequences...

\ CODE REVIEW - MAY NEED TO FIX NOW THAT THIS IS CHANGED

\ To transpose the chord, move the shape
: chf ( -- ) rt@ 1 + rt! shp@ runchord ;
: chb ( -- ) rt@ 1 - rt! shp@ runchord ;


\ Demos
: demo-musiclights
  500 >r 
  32-clear r@ ms
  8 kk     r@ ms
  5 kk     r@ ms
  1 kk     r@ ms
  3 kk     r@ ms
  
  8 kk     r@ ms
  49 light-chord r@ ms
  5 kk     r@ ms
  17 light-chord r@ ms
  1 kk     r@ ms
  13 light-chord r@ ms
  3 kk     r@ ms
  15 light-chord    r@ ms
  
  3 kk     r@ ms
  5 kk     r@ ms
  7 kk     r@ ms
  8 kk     r@ ms
  r> drop 
;


\ ============ INITIALIZE


: init-songs ( -- )
  13 start-speaker
  init-shiftout
  flashlights
  1 buttonsrecord
  s" 3iSongs.fs" included
  well-done!
;


: help-songlist
cr 
cr ." HELP-SONGLIST"
cr ." -------------"
cr ." by Assad Ebrahim - http://www.mathscitech.org/articles"
cr 
cr ." include 3iSonglist.fs  [3iForth.fs 3iCore.fs 3iSound-4x4.fs 3iMusic-scales.fs 3iShift.fs 3iPianoLights.fs]"
cr ." include 3iArpeggiator.fs"
cr ." init: 1 open-com  init-songs"
cr ." playlist"
cr ."      play | watch | pace | learn | showuniqs | transpose | ntranspose "
cr 
cr ." Tools: "
cr ."    replay   e.g. after transpose or ntranspose"
cr ."    ><<   >>>  octave down/up"
cr ."    help-scales   set white keys to diatonic scale for ethnomusic"
cr ."    help-theory   show grammar / words for music theory"
cr
;

: help-theory
cr
cr ." Tools for Music Theory"
cr ." ----------------------"
cr ." Song notation (grammar): 1..8, #,% <<,>> //,///,//3,//// - notes, sharp/flat, octave down/up, rests short/long"
cr ." [writesong]  - converts to piano note grammar, "
cr ." extract-uniqs  show-uniq"
cr ." list-uniqs"

cr
;

help-songlist
help-scales
help-theory

include 3iSongs.fs
include 3iMusic-menu.fs

\ V3 - 2021-05-16---Sun---17:14 - perfected the multi-octave notation with << and >> mapping to [<<] and [>>]
\ V4 - 2021-08-11---Wed---06:28 - MAJOR - perfected integration with 32-key piano lights board and different key signatures (ethnoscales), fixed songlist
\ V5 - 2021-08-20---Fri---07:04 - MAJOR - FEATURE (3iSonglist): Added true teach-and-learn capability to the piano! using poll-note-data (3iArpeggiator) to fetch data
\ V6 - 2021-08-21---Sat---08:30 - MAJOR - PARTIAL FEATURE: 
\ IN PROGRESS: partial transpose capability - runs into trouble juggling handling octave.
\    To Reproduce bug from I have a dream
\    5flats
\    6 7 8 16 2 1 15 5     8 play
\    6 7 8 16 2 1 15 5     8 0 ntranspose-song
\    cstd
\    4 11 12 15 13 16 15 12 16 10   10 play
\ V7 2021-08-24---Tue---01:36 MAJOR REVISION: play/watch/pace/learn first [writesong] to mem then play from mem using notes
\  Song grammer is parsed to write once in the note (fixing scale and octave adjustments)
\  After than transpositions are easy in note format from memory
\ V8 2021-08-25---Wed---05:33 -- fixed learning mode and corrected playlist -- 55 songs
\ V9 2021-08-27---Fri---05:07 -- added extract-uniq algorithm, features: show key-signature (light up all keys needed to play), transpose by 1 semitone or n semitones, annotated all 55 songs with optimal transposition to stay entirely on white keys

\ TODO - 2021-08-27---Fri---05:21 software is now good enough to design hardware user interface and test it out
\ TODO - 2021-09-02---Thu---23:13 add to >>> and ><< for cstd the movement of the black5's, with the check of 0 if 0
